@@ -679,14 +679,21 @@ error_before:
 
 }
 
+#define w_curl_easy_setopt(h, opt, value) \
+	do { \
+		CURLcode rc = curl_easy_setopt(h, opt, value); \
+		if (rc != CURLE_OK) { \
+			LM_ERR("curl_easy_setopt(%d): (%s)\n", opt, curl_easy_strerror(rc)); \
+			goto error; \
+		} \
+	} while (0)
+
 int do_http_op (  const db_con_t* h, const db_key_t* k, const db_op_t* op,
 	     const db_val_t* v, const int n, const db_key_t* c, const int nc,
 	     const db_key_t* uk, const db_val_t* uv, const int nu,
 	     const db_key_t o, const str* custom, db_res_t** r,   int db_op )
 
 {
-
-
 	LM_DBG("Called with db_op=%d\n",db_op);
 
 	static var_str q = {0,0,0};
@@ -804,22 +811,22 @@ int do_http_op (  const db_con_t* h, const db_key_t* k, const db_op_t* op,
 
 	LM_DBG("Sent:%s \n",q.s);
 
-	curl_easy_setopt(conn->handle, CURLOPT_HTTPGET, 1);
-	curl_easy_setopt(conn->handle, CURLOPT_URL, q.s);
+	w_curl_easy_setopt(conn->handle, CURLOPT_HTTPGET, 1L);
+	w_curl_easy_setopt(conn->handle, CURLOPT_URL, q.s);
 
 
-	curl_easy_setopt(conn->handle, CURLOPT_WRITEFUNCTION, receive);
-	curl_easy_setopt(conn->handle, CURLOPT_WRITEDATA, &buff);
+	w_curl_easy_setopt(conn->handle, CURLOPT_WRITEFUNCTION, receive);
+	w_curl_easy_setopt(conn->handle, CURLOPT_WRITEDATA, &buff);
 
 	if( db_op != HTTPDB_QUERY && db_op != HTTPDB_CUSTOM)
 	{
 		LM_DBG("Posted:%s \n",&q.s[middle_poz]);
-		curl_easy_setopt(conn->handle, CURLOPT_POSTFIELDS, &q.s[middle_poz]);
+		w_curl_easy_setopt(conn->handle, CURLOPT_POSTFIELDS, &q.s[middle_poz]);
 
 	}
 
 
-	curl_easy_setopt(conn->handle, CURLOPT_FAILONERROR,1);
+	w_curl_easy_setopt(conn->handle, CURLOPT_FAILONERROR,1L);
 	ret = curl_easy_perform(conn->handle);
 
 
@@ -977,11 +984,8 @@ str url_encode(str s)
 	return rez;
 }
 
-
-
 db_con_t* db_http_init(const str* url)
 {
-
 #define DB_HTTP_BUFF_SIZE 1024
 	char* path;
 	char user_pass[DB_HTTP_BUFF_SIZE];
@@ -1051,19 +1055,19 @@ db_con_t* db_http_init(const str* url)
 	//Disable Expect: 100-continue behavior
 	if (disable_expect) {
 		curl->headers = curl_slist_append(curl->headers, "Expect:");
-		curl_easy_setopt(curl->handle,CURLOPT_HTTPHEADER,curl->headers);
+		w_curl_easy_setopt(curl->handle,CURLOPT_HTTPHEADER,curl->headers);
 	}
 
-	curl_easy_setopt(curl->handle,CURLOPT_SSL_VERIFYPEER,0);
-	curl_easy_setopt(curl->handle,CURLOPT_SSL_VERIFYHOST,0);
+	w_curl_easy_setopt(curl->handle,CURLOPT_SSL_VERIFYPEER,0L);
+	w_curl_easy_setopt(curl->handle,CURLOPT_SSL_VERIFYHOST,0L);
 
-	curl_easy_setopt(curl->handle,CURLOPT_USERPWD,user_pass);
-	curl_easy_setopt(curl->handle,CURLOPT_HTTPAUTH,CURLAUTH_ANY);
+	w_curl_easy_setopt(curl->handle,CURLOPT_USERPWD,user_pass);
+	w_curl_easy_setopt(curl->handle,CURLOPT_HTTPAUTH,CURLAUTH_ANY);
 
-	curl_easy_setopt(curl->handle,CURLOPT_ERRORBUFFER,error_buffer);
+	w_curl_easy_setopt(curl->handle,CURLOPT_ERRORBUFFER,error_buffer);
 #if LIBCURL_VERSION_NUM >= 0x071002
 	LM_DBG("timeout set to %d", db_http_timeout);
-	curl_easy_setopt(curl->handle,CURLOPT_TIMEOUT_MS,db_http_timeout);
+	w_curl_easy_setopt(curl->handle,CURLOPT_TIMEOUT_MS,(long)db_http_timeout);
 #endif
 
 	ret = snprintf(path, DB_HTTP_BUFF_SIZE, "http");
@@ -1128,6 +1132,7 @@ db_con_t* db_http_init(const str* url)
 		next_state[OUT][i] = OUT;
 
 	for( i=0 ; i< 256;i++)
+		/* coverity[copy_paste_error: FALSE] - it should be OUT */
 		next_state[ESC][i] = OUT;
 
 	next_state[ OUT ][ (int)quote_delim ] = IN;
