@@ -87,6 +87,7 @@ static int save_flags_fixup(void** param);
 static int save_flags_fixup_free(void** param);
 static int lookup_flags_fixup(void** param);
 static int lookup_flags_fixup_free(void** param);
+static int bflag_fixup(void** param);
 
 /*! \brief Functions */
 static int add_sock_hdr(struct sip_msg* msg, str *str);
@@ -154,7 +155,8 @@ static const cmd_export_t cmds[] = {
 		{CMD_PARAM_STR, 0, 0},
 		{CMD_PARAM_STR|CMD_PARAM_OPT,0,0},
 		{CMD_PARAM_STR|CMD_PARAM_OPT,0,0},
-		{CMD_PARAM_STR|CMD_PARAM_OPT,0,0}, {0,0,0}},
+		{CMD_PARAM_STR|CMD_PARAM_OPT,0,0},
+		{CMD_PARAM_STR|CMD_PARAM_OPT, bflag_fixup,0}, {0,0,0}},
 		REQUEST_ROUTE|ONREPLY_ROUTE},
 	{"remove_ip_port", (cmd_function)_remove_ip_port, {
 		{CMD_PARAM_STR, 0, 0},
@@ -429,6 +431,21 @@ static int domain_fixup(void** param)
 	return 0;
 }
 
+
+static int bflag_fixup(void** param)
+{
+	unsigned int mask;
+
+	if ((mask=fixup_flag( FLAG_TYPE_BRANCH, (str*)*param))==NAMED_FLAG_ERROR)
+		return E_UNSPEC;
+
+	*param = (void*)(unsigned long)mask;
+
+	return 0;
+}
+
+
+
 static int save_flags_fixup(void **param)
 {
 	struct save_flags default_flags;
@@ -469,7 +486,7 @@ static void mod_destroy(void)
 
 static int add_sock_hdr(struct sip_msg* msg, str *hdr_name)
 {
-	struct socket_info* si;
+	const struct socket_info* si;
 	struct lump* anchor;
 	str hdr;
 	char *p;
