@@ -60,6 +60,8 @@
 
 #include "test/unit_tests.h"
 
+#include "libgen.h"
+
 struct sr_module* modules=0;
 
 #ifdef STATIC_EXEC
@@ -404,12 +406,48 @@ void add_mpath(const char *new_mpath)
 	nmpath->buf[nmpath->len] = '\0';
 }
 
+static struct {
+ char *module;
+ char *name;
+ unsigned int flags;
+} module_warnings[] = {
+	{ "rabbitmq", "'rabbitmq' module has been dropped - please use 'event_rabbitmq' instead!", MOD_WARN_EXIT },
+	{ "event_route", "'event_route' module has been integrated in core file. You no longer need to load the module.", MOD_WARN_SKIP }
+};
+
 /* returns 0 on success , <0 on error */
 int load_module(char* name)
 {
 	int i_tmp, len;
 	struct stat statf;
 	struct mpath *mp;
+	int module_warnings_len;
+	char *base_name;
+
+	base_name = basename(name);
+	len = strlen(base_name);
+	if (strstr(base_name, ".so"))
+		len -= 3;
+
+	module_warnings_len = sizeof(module_warnings) / sizeof(module_warnings[0]);
+
+	for (i_tmp = 0; i_tmp < module_warnings_len; i_tmp++) {
+		if (strncmp(base_name, module_warnings[i_tmp].module, len) == 0) {
+			switch (module_warnings[i_tmp].flags)
+			{
+				case MOD_WARN_EXIT:
+					LM_ERR("%s\n", module_warnings[i_tmp].name);
+					return -1;
+
+				case MOD_WARN_SKIP:
+					LM_WARN("%s\n", module_warnings[i_tmp].name);
+					return 0;
+
+				default:
+					break;
+			}
+		}
+	}
 
 	/* if this is a static module, load it directly */
 	if (load_static_module(name) == 0)
@@ -640,7 +678,7 @@ static int init_mod_child( struct sr_module* m, int rank, char *type,
 	if (!skip_others && init_mod_child(m->next, rank, type, 0) != 0)
 		return -1;
 
-	for (dep = m->sr_deps_init; dep; dep = dep->next)
+	for (dep = m->sr_deps_cinit; dep; dep = dep->next)
 		if (!dep->mod->init_child_done)
 			if (init_mod_child(dep->mod, rank, type, 1) != 0)
 				return -1;
